#pragma once

#include <algorithm>

class Win32Allocator
{
public:
	static void * Allocate(size_t nBytes) throw()
	{
		return ::HeapAlloc( Win32Allocator::s_hHeap, HEAP_ZERO_MEMORY, nBytes );
	}

	static void * Reallocate(void * p, size_t nBytes) throw()
	{
		return ::HeapReAlloc( Win32Allocator::s_hHeap, HEAP_ZERO_MEMORY, p, nBytes );
	}

	static void Free(void * p) throw()
	{
		::HeapFree( Win32Allocator::s_hHeap, 0, p );
	}

private:
	static HANDLE s_hHeap;
};

__declspec(selectany) HANDLE Win32Allocator::s_hHeap = ::GetProcessHeap();

template < typename T >
class CWin32HeapPtr : private ATL::CHeapPtr< T, Win32Allocator >
{
public:
	CWin32HeapPtr() throw()
		: CHeapPtr< T, Win32Allocator >(),
		m_dwWritten( 0 ),
		m_dwTotal( 0 )
	{
	}

	CWin32HeapPtr(CWin32HeapPtr< T > & p) throw()
		: CHeapPtr< T, Win32Allocator >( p ),
		m_dwWritten( p.m_dwWritten ),
		m_dwTotal( p.m_dwTotal )
	{
	}

	CWin32HeapPtr< T > & operator = (CWin32HeapPtr< T > & p) throw()
	{
		CHeapPtr< T, Win32Allocator >::operator = ( p );
		m_dwWritten = p.m_dwWritten;
		m_dwTotal = p.m_dwTotal;

		return *this;
	}

	bool Allocate(size_t nElements = 1) throw()
	{
		if (CHeapPtr< T, Win32Allocator >::Allocate( nElements ))
		{
			m_dwWritten = 0;
			m_dwTotal = nElements * sizeof(T);
			return true;
		}
		else
			return false;
	}

	bool Reallocate(size_t nElements) throw()
	{
		if (CHeapPtr< T, Win32Allocator >::Reallocate( nElements ))
		{
			m_dwTotal = nElements * sizeof(T);
			if (m_dwWritten > m_dwTotal)
			{
				ATLASSERT(false);  // cannot shrink when you've written more
				m_dwWritten = m_dwTotal;
			}
			return true;
		}
		else
			return false;
	}

	bool AllocateBytes(size_t nBytes) throw()
	{
		if (CHeapPtr< T, Win32Allocator >::AllocateBytes( nBytes ))
		{
			m_dwWritten = 0;
			m_dwTotal = nBytes;
			return true;
		}
		else
			return false;
	}

	bool ReallocateBytes(size_t nBytes) throw()
	{
		if (CHeapPtr< T, Win32Allocator >::ReallocateBytes( nBytes ))
		{
			m_dwTotal = nBytes;
			if (m_dwWritten > m_dwTotal)
			{
				ATLASSERT(false);  // cannot shrink when you've written more
				m_dwWritten = m_dwTotal;
			}
			return true;
		}
		else
			return false;
	}

	void Free() throw()
	{
		CHeapPtr< T, Win32Allocator >::Free();
		m_dwWritten = m_dwTotal = 0;
	}

	operator T * () const throw()
	{
		return CHeapPtr< T, Win32Allocator >::operator T * ();
	}

	T * operator -> () const throw()
	{
		return CHeapPtr< T, Win32Allocator >::operator -> ();
	}

	T ** operator & () throw()
	{
		return CHeapPtr< T, Win32Allocator >::operator & ();
	}

	bool Add(T const * ptr, size_t nElements)
	{
		ATLASSERT(m_pData != NULL);

		if (m_dwWritten + nElements * sizeof(T) > m_dwTotal)
			if (!ReallocateBytes( m_dwWritten + nElements * sizeof(T) ))
				return false;

		std::copy( ptr, ptr + nElements, m_pData + m_dwWritten );
		m_dwWritten += nElements * sizeof(T);

		return true;
	}

	size_t GetWrittenBytes() const
	{
		return m_dwWritten;
	}

	size_t GetTotalBytes() const
	{
		return m_dwTotal;
	}

private:
	size_t m_dwWritten;
	size_t m_dwTotal;
};
